/*
 * bldc.c
 *
 *  Created on: 1 Feb 2025
 *      Author: user
 */
#include "bldc.h"
#include "board.h"
#include "peripherals.h"
#include "pin_mux.h"
#include "clock_config.h"
#include "MK66F18.h"
#include "fsl_debug_console.h"

// Global Variables
volatile bool free_running;
volatile int wait_counter;
volatile bool waiting_for_falling_edge;
volatile int threshold = 300;

volatile uint8_t current_phase = 0;
volatile uint8_t current_duty_cycle = START_DUTY_CYCLE;

// Lookup Tables
// GPIO pin masks for enabling motor phases
uint32_t motor_enable_gpioa_mask[] = {
    BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK | BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK | BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK | BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK | BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK | BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK | BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK
};

// FTM (FlexTimer Module) channels for each motor phase
uint32_t motor_direction_ftm_channel[] = {
    FTM0_MOTOR_A_FTM0_CH0_CHANNEL,
    FTM0_MOTOR_A_FTM0_CH0_CHANNEL,
    FTM0_MOTOR_C_FTM0_CH2_CHANNEL,
    FTM0_MOTOR_C_FTM0_CH2_CHANNEL,
    FTM0_MOTOR_B_FTM0_CH1_CHANNEL,
    FTM0_MOTOR_B_FTM0_CH1_CHANNEL
};

// ADC channels for sensing back-EMF on the inactive phase
uint8_t back_emf_adc_channel_number[] = {
    ADC0_MOTOR_BEMF10_B_IRQ_CHANNEL,
    ADC0_MOTOR_BEMF10_C_IRQ_CHANNEL,
    ADC0_MOTOR_BEMF10_A_IRQ_CHANNEL,
    ADC0_MOTOR_BEMF10_B_IRQ_CHANNEL,
    ADC0_MOTOR_BEMF10_C_IRQ_CHANNEL,
    ADC0_MOTOR_BEMF10_A_IRQ_CHANNEL
};

// Function Definitions
// Disables all motor phases and sets PWM duty cycle to 0%
int configure_off(void) {
    uint8_t success = 0;

    GPIO_PortClear(BOARD_MOTOR_EN_A_PTA6_GPIO, BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK);
    GPIO_PortClear(BOARD_MOTOR_EN_B_PTA7_GPIO, BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK);
    GPIO_PortClear(BOARD_MOTOR_EN_C_PTA1_GPIO, BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK);

    if (FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, FTM0_MOTOR_A_FTM0_CH0_CHANNEL,
    				kFTM_EdgeAlignedPwm, 0) == kStatus_Success &&
        FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, FTM0_MOTOR_B_FTM0_CH1_CHANNEL,
        			kFTM_EdgeAlignedPwm, 0) == kStatus_Success &&
        FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, FTM0_MOTOR_C_FTM0_CH2_CHANNEL,
        			kFTM_EdgeAlignedPwm, 0) == kStatus_Success) {
        success = 1;
    }

    return success;
}

// Configures motor phase and PWM duty cycle
uint8_t reconfigure_for_phase(uint8_t phase, uint8_t duty_cycle) {
    uint8_t success = 0;

    if (configure_off()) {
        if (FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, motor_direction_ftm_channel[phase],
        						kFTM_EdgeAlignedPwm,
								duty_cycle) == kStatus_Success) {
            FTM_SetSoftwareTrigger(FTM0_PERIPHERAL, true);
            GPIO_PortSet(GPIOA, motor_enable_gpioa_mask[phase]);
            success = 1;
        } else {
            printf("Error: Failed to configure duty cycle for phase %d\n", phase);
        }
    }

    return success;
}

// Advances to the next commutation phase
void commute(void) {
    current_phase = (current_phase + 1) % 6;
    reconfigure_for_phase(current_phase, current_duty_cycle);
}

// Interrupt handler for back-EMF detection and commutation control
void motor_bemf_irq_handler(void) {
    uint32_t result_values[2] = {0};

    for (int i = 0; i < 2; i++) {
        uint32_t status = ADC16_GetChannelStatusFlags(ADC0_PERIPHERAL, i);
        if (status == kADC16_ChannelConversionDoneFlag) {
            result_values[i] = ADC16_GetChannelConversionValue(ADC0_PERIPHERAL, i);
        }
    }

    if (free_running) {
        if (wait_counter < 20) {
            waiting_for_falling_edge = (result_values[0] >= threshold);
            wait_counter++;
        } else {
            if ((waiting_for_falling_edge && result_values[0] < threshold) ||
            		(!waiting_for_falling_edge && result_values[0] > threshold)) {
                commute();
                wait_counter = 0;
                waiting_for_falling_edge = !waiting_for_falling_edge;
            }
        }
    }

    ADC16_SetChannelConfig(ADC0_PERIPHERAL, ADC0_CH0_CONTROL_GROUP,
    		&ADC0_channelsConfig[back_emf_adc_channel_number[current_phase]]);

#if defined _CORTEX_M && (_CORTEX_M == 4U)
    __DSB();
#endif
}

// Initializes the motor system
void bldc_motor_init(void) {
    current_duty_cycle = START_DUTY_CYCLE;
    printf("Motor Initialized\n");
}

// Sets motor speed based on percentage input
void bldc_motor_set_speed(int percentage) {
    if (percentage < 1) {
        bldc_motor_stop();
    } else {
        current_duty_cycle = 12 + (50 * (percentage / 100));
    }
}

// Stops the motor operation
void bldc_motor_stop(void) {
    free_running = false;
    configure_off();
}

// Forces initial commutations to start motor before BEMF control takes over
void bldc_force_commute_start(void) {
    bldc_motor_init();
    commute();
    int t = 90000;
    for (int j = 10; j <= 35; j += 5) {
        for (int i = 0; i < 30; i++) {
            commute();
            for (int k = 0; k < t; k++) {}
        }
        t -= 10000;
        bldc_motor_set_speed(j);
    }

    free_running = true;
    ADC16_SetChannelConfig(ADC0_PERIPHERAL, ADC0_CH0_CONTROL_GROUP,
                           &ADC0_channelsConfig[back_emf_adc_channel_number[current_phase]]);
}

